@@ -25,6 +25,9 @@
 #include <login_cap.h>
 #endif /* HAVE_LOGIN_CAP_H */
 #endif /* HAVE_SETUSERCONTEXT */
+#ifdef HAVE_OPENSSL_RAND_H
+#include <openssl/rand.h>
+#endif
 
 #include <assert.h>
 #include <ctype.h>
@@ -41,7 +44,9 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#ifdef HAVE_IFADDRS_H
 #include <ifaddrs.h>
+#endif
 
 #include "nsd.h"
 #include "options.h"
@@ -145,135 +150,6 @@ version(void)
 	exit(0);
 }
 
-#ifdef HAVE_GETIFADDRS
-static void
-resolve_ifa_name(struct ifaddrs *ifas, const char *search_ifa, char ***ip_addresses, size_t *ip_addresses_size)
-{
-	struct ifaddrs *ifa;
-	size_t last_ip_addresses_size = *ip_addresses_size;
-
-	for(ifa = ifas; ifa != NULL; ifa = ifa->ifa_next) {
-		sa_family_t family;
-		const char* atsign;
-#ifdef INET6      /* |   address ip    | % |  ifa name  | @ |  port  | nul */
-		char addr_buf[INET6_ADDRSTRLEN + 1 + IF_NAMESIZE + 1 + 16 + 1];
-#else
-		char addr_buf[INET_ADDRSTRLEN + 1 + 16 + 1];
-#endif
-
-		if((atsign=strrchr(search_ifa, '@')) != NULL) {
-			if(strlen(ifa->ifa_name) != (size_t)(atsign-search_ifa)
-			   || strncmp(ifa->ifa_name, search_ifa,
-			   atsign-search_ifa) != 0)
-				continue;
-		} else {
-			if(strcmp(ifa->ifa_name, search_ifa) != 0)
-				continue;
-			atsign = "";
-		}
-
-		if(ifa->ifa_addr == NULL)
-			continue;
-
-		family = ifa->ifa_addr->sa_family;
-		if(family == AF_INET) {
-			char a4[INET_ADDRSTRLEN + 1];
-			struct sockaddr_in *in4 = (struct sockaddr_in *)
-				ifa->ifa_addr;
-			if(!inet_ntop(family, &in4->sin_addr, a4, sizeof(a4)))
-				error("inet_ntop");
-			snprintf(addr_buf, sizeof(addr_buf), "%s%s",
-				a4, atsign);
-		}
-#ifdef INET6
-		else if(family == AF_INET6) {
-			struct sockaddr_in6 *in6 = (struct sockaddr_in6 *)
-				ifa->ifa_addr;
-			char a6[INET6_ADDRSTRLEN + 1];
-			char if_index_name[IF_NAMESIZE + 1];
-			if_index_name[0] = 0;
-			if(!inet_ntop(family, &in6->sin6_addr, a6, sizeof(a6)))
-				error("inet_ntop");
-			if_indextoname(in6->sin6_scope_id,
-				(char *)if_index_name);
-			if (strlen(if_index_name) != 0) {
-				snprintf(addr_buf, sizeof(addr_buf),
-					"%s%%%s%s", a6, if_index_name, atsign);
-			} else {
-				snprintf(addr_buf, sizeof(addr_buf), "%s%s",
-					a6, atsign);
-			}
-		}
-#endif
-		else {
-			continue;
-		}
-		VERBOSITY(4, (LOG_INFO, "interface %s has address %s",
-			search_ifa, addr_buf));
-
-		*ip_addresses = xrealloc(*ip_addresses, sizeof(char *) * (*ip_addresses_size + 1));
-		(*ip_addresses)[*ip_addresses_size] = xstrdup(addr_buf);
-		(*ip_addresses_size)++;
-	}
-
-	if (*ip_addresses_size == last_ip_addresses_size) {
-		*ip_addresses = xrealloc(*ip_addresses, sizeof(char *) * (*ip_addresses_size + 1));
-		(*ip_addresses)[*ip_addresses_size] = xstrdup(search_ifa);
-		(*ip_addresses_size)++;
-	}
-}
-#endif /* HAVE_GETIFADDRS */
-
-static void
-resolve_interface_names(struct nsd_options* options)
-{
-#ifdef HAVE_GETIFADDRS
-	struct ifaddrs *addrs;
-	struct ip_address_option *ip_addr;
-	struct ip_address_option *last = NULL;
-	struct ip_address_option *first = NULL;
-
-	if(getifaddrs(&addrs) == -1)
-		  error("failed to list interfaces");
-
-	/* replace the list of ip_adresses with a new list where the
-	 * interface names are replaced with their ip-address strings
-	 * from getifaddrs.  An interface can have several addresses. */
-	for(ip_addr = options->ip_addresses; ip_addr; ip_addr = ip_addr->next) {
-		char **ip_addresses = NULL;
-		size_t ip_addresses_size = 0, i;
-		resolve_ifa_name(addrs, ip_addr->address, &ip_addresses,
-			&ip_addresses_size);
-
-		for (i = 0; i < ip_addresses_size; i++) {
-			struct ip_address_option *current;
-			/* this copies the range_option, dev, and fib from
-			 * the original ip_address option to the new ones
-			 * with the addresses spelled out by resolve_ifa_name*/
-			current = region_alloc_init(options->region, ip_addr,
-				sizeof(*ip_addr));
-			current->address = region_strdup(options->region,
-				ip_addresses[i]);
-			current->next = NULL;
-			free(ip_addresses[i]);
-
-			if(first == NULL) {
-				first = current;
-			} else {
-				last->next = current;
-			}
-			last = current;
-		}
-		free(ip_addresses);
-	}
-
-	freeifaddrs(addrs);
-	options->ip_addresses = first;
-#else
-	(void)options;
-#endif /* HAVE_GETIFADDRS */
-}
-
 static void
 copyaddrinfo(struct nsd_addrinfo *dest, struct addrinfo *src)
 {
@@ -387,7 +263,6 @@ figure_default_sockets(
 	const char *udp_port, const char *tcp_port,
 	const struct addrinfo *hints)
 {
-	int r;
 	size_t i = 0, n = 1;
 	struct addrinfo ai[2] = { *hints, *hints };
 
@@ -428,6 +303,7 @@ figure_default_sockets(
 		 * automatically mapped to our IPv6 socket.
 		 */
 #ifdef IPV6_V6ONLY
+		int r;
 		struct addrinfo *addrs[2] = { NULL, NULL };
 
 		if((r = getaddrinfo(NULL, udp_port, &ai[0], &addrs[0])) == 0 &&
@@ -465,6 +341,7 @@ figure_default_sockets(
 	figure_socket_servers(&(*tcp)[i], NULL);
 }
 
+#ifdef HAVE_GETIFADDRS
 static int
 find_device(
 	struct nsd_socket *sock,
@@ -513,6 +390,7 @@ find_device(
 
 	return 0;
 }
+#endif /* HAVE_GETIFADDRS */
 
 static void
 figure_sockets(
@@ -524,7 +402,9 @@ figure_sockets(
 	size_t i = 0;
 	struct addrinfo ai = *hints;
 	struct ip_address_option *ip;
+#ifdef HAVE_GETIFADDRS
 	struct ifaddrs *ifa = NULL;
+#endif
 	int bind_device = 0;
 
 	if(!ips) {
@@ -539,9 +419,11 @@ figure_sockets(
 		bind_device |= (ip->dev != 0);
 	}
 
+#ifdef HAVE_GETIFADDRS
 	if(bind_device && getifaddrs(&ifa) == -1) {
 		error("getifaddrs failed: %s", strerror(errno));
 	}
+#endif
 
 	*udp = xalloc_zero((*ifs + 1) * sizeof(struct nsd_socket));
 	*tcp = xalloc_zero((*ifs + 1) * sizeof(struct nsd_socket));
@@ -560,6 +442,7 @@ figure_sockets(
 			(*udp)[i].fib = ip->fib;
 			(*tcp)[i].fib = ip->fib;
 		}
+#ifdef HAVE_GETIFADDRS
 		if(ip->dev != 0) {
 			(*udp)[i].flags |= NSD_BIND_DEVICE;
 			(*tcp)[i].flags |= NSD_BIND_DEVICE;
@@ -570,13 +453,16 @@ figure_sockets(
 				      ip->address);
 			}
 		}
+#endif
 	}
 
 	assert(i == *ifs);
 
+#ifdef HAVE_GETIFADDRS
 	if(ifa != NULL) {
 		freeifaddrs(ifa);
 	}
+#endif
 }
 
 /* print server affinity for given socket. "*" if socket has no affinity with
@@ -668,12 +554,12 @@ print_sockets(
 
 	for(i = 0; i < ifs; i++) {
 		assert(udp[i].servers->size == servercnt);
-		addrport2str(&udp[i].addr.ai_addr, sockbuf, sizeof(sockbuf));
+		addrport2str((void*)&udp[i].addr.ai_addr, sockbuf, sizeof(sockbuf));
 		print_socket_servers(&udp[i], serverbuf, serverbufsz);
 		nsd_bitset_or(servers, servers, udp[i].servers);
 		VERBOSITY(3, (LOG_NOTICE, fmt, sockbuf, "udp", serverbuf));
 		assert(tcp[i].servers->size == servercnt);
-		addrport2str(&tcp[i].addr.ai_addr, sockbuf, sizeof(sockbuf));
+		addrport2str((void*)&tcp[i].addr.ai_addr, sockbuf, sizeof(sockbuf));
 		print_socket_servers(&tcp[i], serverbuf, serverbufsz);
 		nsd_bitset_or(servers, servers, tcp[i].servers);
 		VERBOSITY(3, (LOG_NOTICE, fmt, sockbuf, "tcp", serverbuf));
@@ -945,6 +831,40 @@ bind8_stats (struct nsd *nsd)
 }
 #endif /* BIND8_STATS */
 
+static
+int cookie_secret_file_read(nsd_type* nsd) {
+	char secret[NSD_COOKIE_SECRET_SIZE * 2 + 2/*'\n' and '\0'*/];
+	char const* file = nsd->options->cookie_secret_file;
+	FILE* f;
+	int corrupt = 0;
+	size_t count;
+
+	assert( nsd->options->cookie_secret_file != NULL );
+	f = fopen(file, "r");
+	/* a non-existing cookie file is not an error */
+	if( f == NULL ) { return errno != EPERM; }
+	/* cookie secret file exists and is readable */
+	nsd->cookie_count = 0;
+	for( count = 0; count < NSD_COOKIE_HISTORY_SIZE; count++ ) {
+		size_t secret_len = 0;
+		ssize_t decoded_len = 0;
+		if( fgets(secret, sizeof(secret), f) == NULL ) { break; }
+		secret_len = strlen(secret);
+		if( secret_len == 0 ) { break; }
+		assert( secret_len <= sizeof(secret) );
+		secret_len = secret[secret_len - 1] == '\n' ? secret_len - 1 : secret_len;
+		if( secret_len != NSD_COOKIE_SECRET_SIZE * 2 ) { corrupt++; break; }
+		/* needed for `hex_pton`; stripping potential `\n` */
+		secret[secret_len] = '\0';
+		decoded_len = hex_pton(secret, nsd->cookie_secrets[count].cookie_secret,
+		                       NSD_COOKIE_SECRET_SIZE);
+		if( decoded_len != NSD_COOKIE_SECRET_SIZE ) { corrupt++; break; }
+		nsd->cookie_count++;
+	}
+	fclose(f);
+	return corrupt == 0;
+}
+
 extern char *optarg;
 extern int optind;
 
@@ -986,12 +906,15 @@ main(int argc, char *argv[])
 	nsd.chrootdir	= 0;
 	nsd.nsid 	= NULL;
 	nsd.nsid_len 	= 0;
+	nsd.cookie_count = 0;
 
 	nsd.child_count = 0;
 	nsd.maximum_tcp_count = 0;
 	nsd.current_tcp_count = 0;
 	nsd.file_rotation_ok = 0;
 
+	nsd.do_answer_cookie = 1;
+
 	/* Set up our default identity to gethostname(2) */
 	if (gethostname(hostname, MAXHOSTNAMELEN) == 0) {
 		nsd.identity = hostname;
@@ -1268,6 +1191,36 @@ main(int argc, char *argv[])
 #endif /* IPV6 MTU) */
 #endif /* defined(INET6) */
 
+	nsd.do_answer_cookie = nsd.options->answer_cookie;
+	if (nsd.cookie_count > 0)
+		; /* pass */
+
+	else if (nsd.options->cookie_secret) {
+		ssize_t len = hex_pton(nsd.options->cookie_secret,
+			nsd.cookie_secrets[0].cookie_secret, NSD_COOKIE_SECRET_SIZE);
+		if (len != NSD_COOKIE_SECRET_SIZE ) {
+			error("A cookie secret must be a "
+			      "128 bit hex string");
+		}
+		nsd.cookie_count = 1;
+	} else {
+		size_t j;
+		size_t const cookie_secret_len = NSD_COOKIE_SECRET_SIZE;
+		/* Calculate a new random secret */
+		srandom(getpid() ^ time(NULL));
+
+		for( j = 0; j < NSD_COOKIE_HISTORY_SIZE; j++) {
+#if defined(HAVE_SSL)
+			if (!RAND_status()
+			    || !RAND_bytes(nsd.cookie_secrets[j].cookie_secret, cookie_secret_len))
+#endif
+			for (i = 0; i < cookie_secret_len; i++)
+				nsd.cookie_secrets[j].cookie_secret[i] = random_generate(256);
+		}
+		// XXX: all we have is a random cookie, still pretend we have one
+		nsd.cookie_count = 1;
+	}
+
 	if (nsd.nsid_len == 0 && nsd.options->nsid) {
 		if (strlen(nsd.options->nsid) % 2 != 0) {
 			error("the NSID must be a hex string of an even length.");
@@ -1549,6 +1502,11 @@ main(int argc, char *argv[])
 	}
 #endif /* HAVE_SSL */
 
+	if(nsd.options->cookie_secret_file && nsd.options->cookie_secret_file[0]
+	   && !cookie_secret_file_read(&nsd) ) {
+		log_msg(LOG_ERR, "cookie secret file corrupt or not readable");
+	}
+
 	/* Unless we're debugging, fork... */
 	if (!nsd.debug) {
 		int fd;
@@ -1703,13 +1661,6 @@ main(int argc, char *argv[])
 	options_zonestatnames_create(nsd.options);
 	server_zonestat_alloc(&nsd);
 #endif /* USE_ZONE_STATS */
-#ifdef USE_DNSTAP
-	if(nsd.options->dnstap_enable) {
-		nsd.dt_collector = dt_collector_create(&nsd);
-		dt_collector_start(nsd.dt_collector, &nsd);
-	}
-#endif /* USE_DNSTAP */
-
 	if(nsd.server_kind == NSD_SERVER_MAIN) {
 		server_prepare_xfrd(&nsd);
 		/* xfrd forks this before reading database, so it does not get
@@ -1717,6 +1668,12 @@ main(int argc, char *argv[])
 		server_start_xfrd(&nsd, 0, 0);
 		/* close zonelistfile in non-xfrd processes */
 		zone_list_close(nsd.options);
+#ifdef USE_DNSTAP
+		if(nsd.options->dnstap_enable) {
+			nsd.dt_collector = dt_collector_create(&nsd);
+			dt_collector_start(nsd.dt_collector, &nsd);
+		}
+#endif /* USE_DNSTAP */
 	}
 	if (server_prepare(&nsd) != 0) {
 		unlinkpid(nsd.pidfile);
