@@ -540,6 +540,9 @@ answer_chaos(struct nsd *nsd, query_type *q)
 				ANCOUNT_SET(q->packet, ANCOUNT(q->packet) + 1);
 			} else {
 				RCODE_SET(q->packet, RCODE_REFUSE);
+				/* RFC8914 - Extended DNS Errors
+				 * 4.19. Extended DNS Error Code 18 - Prohibited */
+				q->edns.ede = EDE_PROHIBITED;
 			}
 		} else if ((q->qname->name_size == 16
 			    && memcmp(dname_name(q->qname), "\007version\006server", 16) == 0) ||
@@ -556,13 +559,23 @@ answer_chaos(struct nsd *nsd, query_type *q)
 				ANCOUNT_SET(q->packet, ANCOUNT(q->packet) + 1);
 			} else {
 				RCODE_SET(q->packet, RCODE_REFUSE);
+				/* RFC8914 - Extended DNS Errors
+				 * 4.19. Extended DNS Error Code 18 - Prohibited */
+				q->edns.ede = EDE_PROHIBITED;
 			}
 		} else {
 			RCODE_SET(q->packet, RCODE_REFUSE);
+			/* RFC8914 - Extended DNS Errors
+			 * 4.22. Extended DNS Error Code 21 - Not Supported */
+			q->edns.ede = EDE_NOT_SUPPORTED;
+
 		}
 		break;
 	default:
 		RCODE_SET(q->packet, RCODE_REFUSE);
+		/* RFC8914 - Extended DNS Errors
+		 * 4.22. Extended DNS Error Code 21 - Not Supported */
+		q->edns.ede = EDE_NOT_SUPPORTED;
 		break;
 	}
 
@@ -1136,6 +1149,10 @@ answer_authoritative(struct nsd   *nsd,
 		++q->cname_count;
 		if(!newname) { /* newname too long */
 			RCODE_SET(q->packet, RCODE_YXDOMAIN);
+			/* RFC 8914 - Extended DNS Errors
+			 * 4.21. Extended DNS Error Code 0 - Other */
+			ASSIGN_EDE_CODE_AND_STRING_LITERAL(q->edns.ede,
+				EDE_OTHER, "DNAME expansion became too large");
 			return;
 		}
 		DEBUG(DEBUG_QUERY,2, (LOG_INFO, "->result is %s", dname_to_string(newname, NULL)));
@@ -1271,15 +1288,61 @@ answer_lookup_zone(struct nsd *nsd, struct query *q, answer_type *answer,
 	q->zone = domain_find_zone(nsd->db, closest_encloser);
 	if (!q->zone) {
 		/* no zone for this */
-		if(q->cname_count == 0)
+		if(q->cname_count == 0) {
 			RCODE_SET(q->packet, RCODE_REFUSE);
+			/* RFC 8914 - Extended DNS Errors
+			 * 4.21. Extended DNS Error Code 20 - Not Authoritative */
+			q->edns.ede = EDE_NOT_AUTHORITATIVE;
+		}
 		return;
 	}
 	assert(closest_encloser); /* otherwise, no q->zone would be found */
+	if(q->zone->opts && q->zone->opts->pattern
+	&& q->zone->opts->pattern->allow_query) {
+		struct acl_options *why = NULL;
+
+		/* check if it passes acl */
+		if(acl_check_incoming(
+		   q->zone->opts->pattern->allow_query, q, &why) != -1) {
+			assert(why);
+			DEBUG(DEBUG_QUERY,1, (LOG_INFO, "query %s passed acl %s %s",
+				dname_to_string(q->qname, NULL),
+				why->ip_address_spec,
+				why->nokey?"NOKEY":
+				(why->blocked?"BLOCKED":why->key_name)));
+		} else { 
+			if (verbosity >= 2) {
+				char address[128];
+				addr2str(&q->addr, address, sizeof(address));
+				VERBOSITY(2, (LOG_INFO, "query %s from %s refused, %s %s",
+					dname_to_string(q->qname, NULL),
+					address,
+					why ? ( why->nokey    ? "NOKEY"
+					      : why->blocked  ? "BLOCKED"
+					      : why->key_name ) 
+					    : "no acl matches",
+					why?why->ip_address_spec:"."));
+			}
+			/* no zone for this */
+			if(q->cname_count == 0) {
+				RCODE_SET(q->packet, RCODE_REFUSE);
+				/* RFC8914 - Extended DNS Errors
+				 * 4.19. Extended DNS Error Code 18 - Prohibited */
+				q->edns.ede = EDE_PROHIBITED;
+			}
+			return;
+		}
+	}
 	if(!q->zone->apex || !q->zone->soa_rrset) {
 		/* zone is configured but not loaded */
-		if(q->cname_count == 0)
+		if(q->cname_count == 0) {
 			RCODE_SET(q->packet, RCODE_SERVFAIL);
+			/* RFC 8914 - Extended DNS Errors
+			 * 4.15. Extended DNS Error Code 14 - Not Ready */
+			q->edns.ede = EDE_NOT_READY;
+			ASSIGN_EDE_CODE_AND_STRING_LITERAL(q->edns.ede,
+			    EDE_NOT_READY, "Zone is configured but not loaded");
+		}
 		return;
 	}
 
@@ -1316,8 +1379,14 @@ answer_lookup_zone(struct nsd *nsd, struct query *q, answer_type *answer,
 			q->zone = zone;
 			if(!q->zone->apex || !q->zone->soa_rrset) {
 				/* zone is configured but not loaded */
-				if(q->cname_count == 0)
+				if(q->cname_count == 0) {
 					RCODE_SET(q->packet, RCODE_SERVFAIL);
+					/* RFC 8914 - Extended DNS Errors
+					 * 4.15. Extended DNS Error Code 14 - Not Ready */
+					ASSIGN_EDE_CODE_AND_STRING_LITERAL(
+					   q->edns.ede, EDE_NOT_READY,
+					   "Zone is configured but not loaded");
+				}
 				return;
 			}
 		}
@@ -1326,8 +1395,13 @@ answer_lookup_zone(struct nsd *nsd, struct query *q, answer_type *answer,
 	/* see if the zone has expired (for secondary zones) */
 	if(q->zone && q->zone->opts && q->zone->opts->pattern &&
 		q->zone->opts->pattern->request_xfr != 0 && !q->zone->is_ok) {
-		if(q->cname_count == 0)
+		if(q->cname_count == 0) {
 			RCODE_SET(q->packet, RCODE_SERVFAIL);
+			/* RFC 8914 - Extended DNS Errors
+			 * 4.25. Extended DNS Error Code 24 - Invalid Data */
+			ASSIGN_EDE_CODE_AND_STRING_LITERAL(q->edns.ede,
+				EDE_INVALID_DATA, "Zone has expired");
+		}
 		return;
 	}
 
@@ -1350,6 +1424,7 @@ answer_lookup_zone(struct nsd *nsd, struct query *q, answer_type *answer,
 		}
 
 		if (!q->delegation_domain
+		    || !q->delegation_rrset
 		    || (exact && q->qtype == TYPE_DS && closest_encloser == q->delegation_domain))
 		{
 			if (q->qclass == CLASS_ANY) {
@@ -1421,7 +1496,7 @@ query_prepare_response(query_type *q)
  *
  */
 query_state_type
-query_process(query_type *q, nsd_type *nsd)
+query_process(query_type *q, nsd_type *nsd, uint32_t *now_p)
 {
 	/* The query... */
 	nsd_rc_type rc;
@@ -1481,7 +1556,7 @@ query_process(query_type *q, nsd_type *nsd)
 				if(process_edns(nsd, q) == NSD_RC_OK) {
 					int opcode = OPCODE(q->packet);
 					(void)query_error(q, NSD_RC_FORMAT);
-					query_add_optional(q, nsd);
+					query_add_optional(q, nsd, now_p);
 					FLAGS_SET(q->packet, FLAGS(q->packet) & 0x0100U);
 						/* Preserve the RD flag. Clear the rest. */
 					OPCODE_SET(q->packet, opcode);
@@ -1516,22 +1591,17 @@ query_process(query_type *q, nsd_type *nsd)
 	}
 
 	arcount = ARCOUNT(q->packet);
-	if (arcount > 0) {
-		/* According to draft-ietf-dnsext-rfc2671bis-edns0-10:
-		 * "The placement flexibility for the OPT RR does not
-		 * override the need for the TSIG or SIG(0) RRs to be
-		 * the last in the additional section whenever they are
-		 * present."
-		 * So we should not have to check for TSIG RR before
-		 * OPT RR. Keep the code for backwards compatibility.
-		 */
-
-		/* see if tsig is before edns record */
-		if (!tsig_parse_rr(&q->tsig, q->packet))
-			return query_formerr(q, nsd);
-		if(q->tsig.status != TSIG_NOT_PRESENT)
-			--arcount;
-	}
+	/* A TSIG RR is not allowed before the EDNS OPT RR.
+	 * In RFC6891 (about EDNS) it says:
+	 * "The placement flexibility for the OPT RR does not
+	 * override the need for the TSIG or SIG(0) RRs to be
+	 * the last in the additional section whenever they are
+	 * present."
+	 * And in RFC8945 (about TSIG) it says:
+	 * "If multiple TSIG records are detected or a TSIG record is
+	 * present in any other position, the DNS message is dropped
+	 * and a response with RCODE 1 (FORMERR) MUST be returned."
+	 */
 	/* See if there is an OPT RR. */
 	if (arcount > 0) {
 		if (edns_parse_record(&q->edns, q->packet, q, nsd))
@@ -1583,16 +1653,21 @@ query_process(query_type *q, nsd_type *nsd)
 		return QUERY_PROCESSED;
 	}
 
+	if (q->edns.cookie_status == COOKIE_UNVERIFIED)
+		cookie_verify(q, nsd, now_p);
+
 	query_prepare_response(q);
 
 	if (q->qclass != CLASS_IN && q->qclass != CLASS_ANY) {
 		if (q->qclass == CLASS_CH) {
 			return answer_chaos(nsd, q);
 		} else {
-			return query_error(q, NSD_RC_REFUSE);
+			/* RFC8914 - Extended DNS Errors
+			 * 4.22. Extended DNS Error Code 21 - Not Supported */
+			q->edns.ede = EDE_NOT_SUPPORTED;
+			return query_error(q, RCODE_REFUSE);
 		}
 	}
-
 	query_state = answer_axfr_ixfr(nsd, q);
 	if (query_state == QUERY_PROCESSED || query_state == QUERY_IN_AXFR) {
 		return query_state;
@@ -1608,7 +1683,7 @@ query_process(query_type *q, nsd_type *nsd)
 }
 
 void
-query_add_optional(query_type *q, nsd_type *nsd)
+query_add_optional(query_type *q, nsd_type *nsd, uint32_t *now_p)
 {
 	struct edns_data *edns = &nsd->edns_ipv4;
 #if defined(INET6)
@@ -1626,6 +1701,14 @@ query_add_optional(query_type *q, nsd_type *nsd)
 		if (q->edns.dnssec_ok)	edns->ok[7] = 0x80;
 		else			edns->ok[7] = 0x00;
 		buffer_write(q->packet, edns->ok, OPT_LEN);
+
+		/* Add Extended DNS Error (RFC8914)
+		 * to verify that we stay in bounds */
+		if (q->edns.ede >= 0)
+			q->edns.opt_reserved_space +=
+				6 + ( q->edns.ede_text_len
+			            ? q->edns.ede_text_len : 0);
+
 		if(q->edns.opt_reserved_space == 0 || !buffer_available(
 			q->packet, 2+q->edns.opt_reserved_space)) {
 			/* fill with NULLs */
@@ -1640,6 +1723,29 @@ query_add_optional(query_type *q, nsd_type *nsd)
 				/* nsid payload */
 				buffer_write(q->packet, nsd->nsid, nsd->nsid_len);
 			}
+			if(q->edns.cookie_status != COOKIE_NOT_PRESENT) {
+				/* cookie opt header */
+				buffer_write(q->packet, edns->cookie, OPT_HDR);
+				/* cookie payload */
+				cookie_create(q, nsd, now_p);
+				buffer_write(q->packet, q->edns.cookie, 24);
+			}
+			/* Append Extended DNS Error (RFC8914) option if needed */
+			if (q->edns.ede >= 0) { /* < 0 means no EDE */
+				/* OPTION-CODE */
+				buffer_write_u16(q->packet, EDE_CODE);
+				/* OPTION-LENGTH */
+				buffer_write_u16(q->packet,
+					2 + ( q->edns.ede_text_len
+					    ? q->edns.ede_text_len : 0));
+				/* INFO-CODE */
+				buffer_write_u16(q->packet, q->edns.ede);
+				/* EXTRA-TEXT */
+				if (q->edns.ede_text_len)
+					buffer_write(q->packet,
+							q->edns.ede_text,
+							q->edns.ede_text_len);
+			}
 		}
 		ARCOUNT_SET(q->packet, ARCOUNT(q->packet) + 1);
 		STATUP(nsd, edns);
