@@ -100,6 +100,27 @@ void rrl_mmap_init(int numch, size_t numbuck, size_t lm, size_t wlm, size_t sm,
 #endif
 }
 
+void rrl_mmap_deinit(void)
+{
+#ifdef HAVE_MMAP
+	size_t i;
+	for(i=0; i<rrl_maps_num; i++) {
+		munmap(rrl_maps[i], sizeof(struct rrl_bucket)*rrl_array_size);
+		rrl_maps[i] = NULL;
+	}
+	free(rrl_maps);
+	rrl_maps = NULL;
+#endif
+}
+
+void rrl_mmap_deinit_keep_mmap(void)
+{
+#ifdef HAVE_MMAP
+	free(rrl_maps);
+	rrl_maps = NULL;
+#endif
+}
+
 void rrl_set_limit(size_t lm, size_t wlm, size_t sm)
 {
 	rrl_ratelimit = lm*2;
@@ -117,6 +138,13 @@ void rrl_init(size_t ch)
 #endif
 }
 
+void rrl_deinit(size_t ch)
+{
+	if(!rrl_maps || ch >= rrl_maps_num)
+		free(rrl_array);
+	rrl_array = NULL;
+}
+
 /** return the source netblock of the query, this is the genuine source
  * for genuine queries and the target for reflected packets */
 static uint64_t rrl_get_source(query_type* query, uint16_t* c2)
@@ -156,7 +184,7 @@ static const char* rrlsource2str(uint64_t s, uint16_t c2)
 		if(!inet_ntop(AF_INET6, &a6, buf, sizeof(buf)))
 			strlcpy(buf, "[ip6 ntop failed]", sizeof(buf));
 		else {
-			static char prefix[4];
+			static char prefix[5];
 			snprintf(prefix, sizeof(prefix), "/%d", rrl_ipv6_prefixlen);
 			strlcat(buf, &prefix[0], sizeof(buf));
 		}
@@ -170,7 +198,7 @@ static const char* rrlsource2str(uint64_t s, uint16_t c2)
 	if(!inet_ntop(AF_INET, &a4, buf, sizeof(buf)))
 		strlcpy(buf, "[ip4 ntop failed]", sizeof(buf));
 	else {
-		static char prefix[4];
+		static char prefix[5];
 		snprintf(prefix, sizeof(prefix), "/%d", rrl_ipv4_prefixlen);
 		strlcat(buf, &prefix[0], sizeof(buf));
 	}
