@@ -1,7 +1,7 @@
 /*
  * tsig-openssl.h -- Interface to OpenSSL for TSIG support.
  *
- * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -22,7 +22,7 @@
  */
 int tsig_openssl_init(region_type *region);
 
-void tsig_openssl_finalize();
+void tsig_openssl_finalize(void);
 
 #endif /* defined(HAVE_SSL) */
 
